require 'record'

class Record
  module Associable
    module ClassMethods
      # Specifies a one-to-one association.
      # @return [void]
      def belongs_to(name, options = {})
        options.each { |key, value| instance_variable_set("@#{key}", value) }
        raise ArgumentError unless /\A\w+\z/ =~ name && /\A\w+\z/ =~ @class_name

        @class_name = "#{self.name.split('::')[0...-1].join('::')}::#{@class_name}"
        @foreign_key = "#{name}_id"

        class_eval <<-END, __FILE__, __LINE__ + 1
          attr_accessor :#{@foreign_key}

          # Returns the associated record.
          # @return [Record, nil]
          def #{name}
            @#{name} ||= #{@class_name}.find(@#{@foreign_key})
          end

          # Sets the associated record.
          # @return [void]
          def #{name}=(#{name})
            @#{name} = #{name}
            @#{@foreign_key} = #{name}.id
          end
        END
      end

      # Specifies a one-to-many association.
      # @return [void]
      def has_many(name, options = {})
        options.each { |key, value| instance_variable_set("@#{key}", value) }
        raise ArgumentError unless /\A\w+\z/ =~ name && /\A\w+\z/ =~ @class_name

        @class_name = "#{self.name.split('::')[0...-1].join('::')}::#{@class_name}"
        @foreign_key = "#{self.name.split('::').last.downcase}_id"

        class_eval <<-END, __FILE__, __LINE__ + 1
          # Returns all associated records.
          # @return [<Record>]
          def #{name}
            records = #{@class_name}.all.select { |record| record.#{@foreign_key} == @id }
            records.instance_variable_set(:@#{@foreign_key}, @id)

            class << records
              # Creates an associated record.
              # @return [Record]
              def create(attributes = {}, &block)
                attributes ||= {}
                attributes.merge!(:#{@foreign_key} => @#{@foreign_key})
                record = #{@class_name}.create(attributes, &block)
                self << record
                record
              end

              # Returns the specified associated record or records.
              # @return [<Record>, Record, nil]
              def find(*args)
                case
                when args[0].is_a?(Array)
                  records = args[0].map { |id| select { |record| record.id == id.to_i } }
                  records.flatten!

                when args.length > 1
                  records = args.map { |id| select { |record| record.id == id.to_i } }
                  records.flatten!

                else
                  records = select { |record| record.id == args[0].to_i }
                  records.first
                end
              end

              # Returns the specified first number of associated records
              # sorted by id.
              # @return [<Record>, Record, nil]
              def first(*args)
                records = sort_by { |record| record.id }
                records.first(*args)
              end

              # Returns the specified last number of associated records
              # sorted by id.
              # @return [<Record>, Record, nil]
              def last(*args)
                records = sort_by { |record| record.id }
                records.last(*args)
              end
            end

            records
          end

          alias_method :destroy_without_#{name}, :destroy

          # Destroys self and all associated records.
          # @return [void]
          def destroy
            records = #{@class_name}.all.select { |record| record.#{@foreign_key} == @id }
            records.each { |record| record.destroy }
            destroy_without_#{name}
          end
        END
      end

      # Specifies a one-to-one association.
      # @return [void]
      def has_one(name, options = {})
        options.each { |key, value| instance_variable_set("@#{key}", value) }
        raise ArgumentError unless /\A\w+\z/ =~ name && /\A\w+\z/ =~ @class_name

        @class_name = "#{self.name.split('::')[0...-1].join('::')}::#{@class_name}"
        @foreign_key = "#{self.name.split('::').last.downcase}_id"

        class_eval <<-END, __FILE__, __LINE__ + 1
          attr_accessor :#{@foreign_key}

          # Returns the associated record.
          # @return [Record, nil]
          def #{name}
            @#{name} ||= begin
              records = #{@class_name}.all.select { |record| record.#{@foreign_key} == @id }
              records.first
            end
          end

          # Sets the associated record.
          # @return [void]
          def #{name}=(#{name})
            @#{name} = #{name}
            #{name}.#{@foreign_key} = @id
          end
        END
      end
    end

    class << self
      def included(base)
        base.__send__(:extend, ClassMethods)
      end
    end
  end
end
