require 'record'

class Child < Record
  belongs_to :parent, :class_name => 'Parent'
end

class Parent < Record
  has_many :children, :class_name => 'Child'
  has_one :child, :class_name => 'Child'
end

describe Record::Associable do
  before(:each) do
    Child.records = []
    Child.records << Child.new(:id => 1, :parent_id => 1)
    Child.records << Child.new(:id => 2, :parent_id => 1)
    Child.records << Child.new(:id => 3, :parent_id => 1)
    Child.records << Child.new(:id => 4, :parent_id => 2)
    Child.records << Child.new(:id => 5, :parent_id => 2)
    Child.records << Child.new(:id => 6, :parent_id => 2)
    Child.records << Child.new(:id => 7, :parent_id => 3)
    Child.records << Child.new(:id => 8, :parent_id => 3)
    Child.records << Child.new(:id => 9, :parent_id => 3)

    Parent.records = []
    Parent.records << Parent.new(:id => 1)
    Parent.records << Parent.new(:id => 2)
    Parent.records << Parent.new(:id => 3)
  end

  let(:child) { Child.records.last }
  let(:parent) { Parent.records.last }
  let(:children) { Child.records.select { |record| record.parent_id == parent.id }}
  let(:children_sorted) { children.sort_by { |record| record.id } }

  context 'when declare a "belongs_to" association' do
    describe '#association' do
      it 'returns the associated record' do
        expect(child.parent).to eq(parent)
      end
    end

    describe '#association=' do
      it 'sets the associated record' do
        child = Child.new(:id => 10)
        child.parent = parent
        expect(child.parent).to eq(parent)
      end
    end
  end

  context 'when declare a "has_many" association' do
    describe '#collection' do
      it 'returns all associated records' do
        expect(parent.children).to match_array(children)
      end
    end

    describe '#collection.create' do
      it 'creates an associated record' do
        record = parent.children.create(:id => 10)
        expect(record).to be_instance_of(Child)
        expect(record).to eq(children.last)
      end
    end

    describe '#collection.find' do
      context 'with an array argument' do
        it 'returns the specified associated record or records' do
          expect(parent.children.find([7, 8, 9])).to match_array(children.first(3))
        end
      end

      context 'with more than one argument' do
        it 'returns the specified associated record or records' do
          expect(parent.children.find(7, 8, 9)).to match_array(children.first(3))
        end
      end

      context 'with one argument' do
        it 'returns the specified associated record' do
          expect(parent.children.find(7)).to eq(children.first)
        end
      end
    end

    describe '#collection.first' do
      context 'with an integer argument' do
        it 'returns the specified first number of associated records sorted by id' do
          expect(parent.children.first(3)).to match_array(children_sorted.first(3))
        end
      end

      context 'with no arguments' do
        it 'returns the first associated record sorted by id' do
          expect(parent.children.first).to eq(children_sorted.first)
        end
      end
    end

    describe '#collection.last' do
      context 'with an integer argument' do
        it 'returns the specified last number of associated records sorted by id' do
          expect(parent.children.last(3)).to match_array(children_sorted.last(3))
        end
      end

      context 'with no arguments' do
        it 'returns the last associated record sorted by id' do
          expect(parent.children.last).to eq(children_sorted.last)
        end
      end
    end
  end

  context 'when declare a "has_one" association' do
    before(:each) do
      Child.records = []
      Child.records << Child.new(:id => 1, :parent_id => 1)
      Child.records << Child.new(:id => 2, :parent_id => 2)
      Child.records << Child.new(:id => 3, :parent_id => 3)

      Parent.records = []
      Parent.records << Parent.new(:id => 1)
      Parent.records << Parent.new(:id => 2)
      Parent.records << Parent.new(:id => 3)
    end

    describe '#association' do
      it 'returns the associated record' do
        expect(parent.child).to eq(child)
      end
    end

    describe '#association=' do
      it 'sets the associated record' do
        parent = Parent.new(:id => 4)
        parent.child = child
        expect(parent.child).to eq(child)
      end
    end
  end
end
